///********************************************************************
///
///     fastNLO_toolkit: fnlo-tk-yodaout
///     Program to read fastNLO tables and write out QCD cross sections
///     with uncertainty in text and YODA format (for use with Rivet)
///
///     For more explanations type:
///     ./fnlo-tk-yodaout -h
///
///     K. Rabbertz, G. Sieber, S. Tyros
///
///********************************************************************

// Precompiler variables for conditional compilation are generated and
// stored automatically in config.h via AC_DEFINE statements in configure.ac.
// To enable conditional compilation, e.g. using HAVE_LIBZ, this config file
// MUST be the very first one to be included with
#include <config.h>

#include <cmath>
#include <cstdlib>
#include <fstream>
#include <iostream>
#include <string>
#include <vector>
#include "fastnlotk/fastNLOAlphas.h"
#include "fastnlotk/fastNLOConstants.h"
#include "fastnlotk/fastNLOLHAPDF.h"
#include "fastnlotk/fastNLOTools.h"
#include "fastnlotk/speaker.h"
#ifdef WITH_YODA
#include "YODA/Scatter2D.h"
#include "YODA/WriterYODA.h"
#endif

//__________________________________________________________________________________________________________________________________
int main(int argc, char** argv) {

   //! --- namespaces
   using namespace std;
   using namespace say;       //! namespace for 'speaker.h'-verbosity levels
   using namespace fastNLO;   //! namespace for fastNLO constants

   //! --- Set initial verbosity level
   // Could set verbosity level directly using enum here, but we'll need fVerbosity anyway for cmd line parsing
   //   SetGlobalVerbosity(WARNING);
   string VerbosityLevel = "WARNING";
   say::Verbosity fVerbosity = toVerbosity()["WARNING"];
   SetGlobalVerbosity(fVerbosity);

   //--- Set verbosity level of table evaluation from command line argument if existing
   if (argc > 11) {
      VerbosityLevel  = (const char*) argv[11];
      //! --- Reset verbosity level from here on
      fVerbosity = toVerbosity()[VerbosityLevel];
      SetGlobalVerbosity(fVerbosity);
   }

   //! --- Parse commmand line
   char buffer[1024];
   string tablename;
   if (argc <= 1) {
      yell << "" << endl;
      yell << _CSEPSC << endl;
      shout["fnlo-tk-yodaout"] << "fastNLO YODA Writer" << endl;
      yell << _SSEPSC << endl;
      error["fnlo-tk-yodaout"] << "No fastNLO table specified!" << endl;
      shout["fnlo-tk-yodaout"] << "For an explanation of command line arguments type:" << endl;
      shout["fnlo-tk-yodaout"] << "./fnlo-tk-yodaout -h" << endl;
      shout["fnlo-tk-yodaout"] << "For version number printout type:" << endl;
      shout["fnlo-tk-yodaout"] << "./fnlo-tk-yodaout -v" << endl;
      yell << _CSEPSC << endl;
      exit(1);
   } else {
      tablename = (const char*) argv[1];
      if (tablename == "-v") {
         fastNLOTools::PrintFastnloVersion();
         return 0;
      } else if (tablename == "-h") {
         // Avoid suppressing usage printouts
         SetGlobalVerbosity(INFO);
      }
      //! --- Print program purpose
      yell << _CSEPSC << endl;
      info["fnlo-tk-yodaout"] << "Program to read fastNLO tables and write out" << endl;
      info["fnlo-tk-yodaout"] << "QCD cross sections with uncertainties in text format and" << endl;
      info["fnlo-tk-yodaout"] << "in YODA format for use with Rivet" << endl;
      info["fnlo-tk-yodaout"] << "(If compiled without YODA support only text printout is given)" << endl;
      infosep << _SSEPSC << endl;
      info["fnlo-tk-yodaout"] << "For an explanation of command line arguments type:" << endl;
      info["fnlo-tk-yodaout"] << "./fnlo-tk-yodaout -h" << endl;
      info["fnlo-tk-yodaout"] << "For version number printout type:" << endl;
      info["fnlo-tk-yodaout"] << "./fnlo-tk-yodaout -v" << endl;
      infosep << _CSEPSC << endl;
      infosep << "" << endl;
      infosep << _CSEPSC << endl;
      //! --- Usage info
      if (tablename == "-h") {
         info["fnlo-tk-yodaout"] << "fastNLO YODA Writer" << endl;
         infosep << _SSEPSC << endl;
         infosep << " #" << endl;
         info["fnlo-tk-yodaout"] << "This program evaluates a fastNLO table and" << endl;
         info["fnlo-tk-yodaout"] << "prints out cross sections with statistical (if available), " << endl;
         info["fnlo-tk-yodaout"] << "scale, or PDF uncertainties in YODA format for use with Rivet." << endl;
         info["fnlo-tk-yodaout"] << "For this to work, the scenario description must contain" << endl;
         info["fnlo-tk-yodaout"] << "the Rivet ID in the form 'RIVET_ID=EXP_YYYY_INSPIREID/Dii-xjj-ykk'," << endl;
         info["fnlo-tk-yodaout"] << "where 'ii', 'jj', and 'kk' indicate the first histogram covered by" << endl;
         info["fnlo-tk-yodaout"] << "this table and the capital letter indicates the plot counter to increase." << endl;
         info["fnlo-tk-yodaout"] << "In case the Rivet ID is missing, it can be added e.g. by using 'fnlo-tk-modify'." << endl;
         man << "" << endl;
         man << "Usage: ./fnlo-tk-yodaout <fastNLOtable.tab.gz> [PDF] [uncertainty]" << endl;
         man << "       Arguments: <> mandatory; [] optional." << endl;
         man << "<fastNLOtable.tab>: Table input file, e.g. fnl1234.tab.gz" << endl;
         man << "[PDF]: PDF set, def. = CT18NNLO" << endl;
         man << "   For LHAPDF6: Specify set names WITHOUT filename extension." << endl;
         man << "   If the PDF set still is not found, then:" << endl;
         man << "   - Check, whether the LHAPDF environment variable is set correctly." << endl;
         man << "   - Specify the PDF set including the absolute path." << endl;
         man << "   - Download the desired PDF set from the LHAPDF web site." << endl;
         man << "[uncertainty]: Uncertainty to show, def. = none" << endl;
         man << "   Alternatives: NN (none, but correct MC sampling average value --> NNPDF PDFs)" << endl;
         man << "                 2P (symmetric 2-point scale factor variation)" << endl;
         man << "                 6P (asymmetric 6-point scale factor variation)" << endl;
         man << "                 HS (symmetric Hessian PDF uncertainty --> ABM, (G)JR PDFs)" << endl;
         man << "                 HA (asymmetric Hessian PDF uncertainty)" << endl;
         man << "                 HP (pairwise asymmetric Hessian PDF uncertainty --> CTEQ|MSTW PDFs)" << endl;
         man << "                 HC (pairwise asymmetric Hessian PDF uncertainty rescaled to CL68 --> CTEQ PDFs)" << endl;
         man << "                 MC (MC sampling PDF uncertainty --> NNPDF PDFs)" << endl;
         man << "                 L6 (LHAPDF6 PDF uncertainty --> LHAPDF6 PDFs)" << endl;
         man << "                 AS (a_s(M_Z) variation uncertainty with GRV evolution)" << endl;
         man << "                 ST (statistical uncertainty of x section calculation, if available)" << endl;
         man << "[order]: Fixed-order precision to use, def. = NLO" << endl;
         man << "   Alternatives: LO, NLO_only, NNLO, NNLO_only (if available)" << endl;
         man << "[norm]: Normalize if applicable, def. = no." << endl;
         man << "   Alternatives: \"yes\" or \"norm\"" << endl;
         man << "[flexscale]: Central scale choice for flex-scale tables." << endl;
         man << "   Default:      \"kScale1\",  i.e. mur=muf=scale1," << endl;
         man << "   Alternatives: \"kScale2\",  i.e. mur=muf=scale2," << endl;
         man << "                 \"scale12\", i.e. mur=scale1, muf=scale2," << endl;
         man << "                 \"scale21\", i.e. mur=scale2, muf=scale1," << endl;
         man << "                 \"kProd\", i.e. mur=muf=scale1*scale2," << endl;
         man << "                 \"kQuadraticSum\", i.e. mur=muf=sqrt(scale1^2+scale2^2)." << endl;
         man << "[sclfac]: Central scale factor (e.g. to change from pT to 2*pT or m12 to m12/2), def. = 1" << endl;
         man << "[np]: Apply nonperturbative corrections if available, def. = no.; alternatives: \"yes\" or \"np\"." << endl;
         man << "[xrescale]: Rescale x axis values, def. = 1" << endl;
         man << "[dxsrescale]: Rescale uncertainty values, def. = 1" << endl;
         man << "[Verbosity]: Set verbosity level of table evaluation [DEBUG,INFO,WARNING,ERROR,SILENT], def. = WARNING" << endl;
         infosep << " #" << endl;
         man << "Use \"_\" to skip changing a default argument." << endl;
         infosep << " #" << endl;
         infosep << _CSEPSC << endl;
         return 0;
      } else {
         shout["fnlo-tk-yodaout"] << "Evaluating table: "  <<  tablename << endl;
      }
   }

   //! --- PDF choice
   string chtmp;
   string PDFFile;
   if (argc > 2) {
      PDFFile = (const char*) argv[2];
   }
   if (argc <= 2 || PDFFile == "_") {
      PDFFile = "CT18NNLO";
      shout["fnlo-tk-yodaout"] << "No PDF set given, taking " << PDFFile << " instead!" << endl;
   } else {
      shout["fnlo-tk-yodaout"] << "Using PDF set   : " << PDFFile << endl;
   }

   //! --- Uncertainty choice
   EScaleUncertaintyStyle eScaleUnc = kScaleNone;
   EPDFUncertaintyStyle   ePDFUnc   = kPDFNone;
   EAsUncertaintyStyle    eAsUnc    = kAsNone;
   EStatUncertaintyStyle   eStatUnc = kStatNone;
   string chunc;
   if (argc > 3) {
      chunc = (const char*) argv[3];
   }
   if (argc <= 3 || chunc == "_") {
      chunc = "none";
      shout["fnlo-tk-yodaout"] << "No request given for uncertainty, none evaluated." << endl;
   } else {
      if ( chunc == "NN" ) {
         shout["fnlo-tk-yodaout"] << "No PDF uncertainty, but correct MC sampling average value as needed for NNPDF." << endl;
      } else if ( chunc == "2P" ) {
         eScaleUnc = kSymmetricTwoPoint;
         shout["fnlo-tk-yodaout"] << "Showing uncertainty from symmetric 2-point scale factor variation." << endl;
      } else if ( chunc == "6P" ) {
         eScaleUnc = kAsymmetricSixPoint;
         shout["fnlo-tk-yodaout"] << "Showing uncertainty from asymmetric 6-point scale factor variation." << endl;
      } else if ( chunc == "HS" ) {
         ePDFUnc = kHessianSymmetric;
         shout["fnlo-tk-yodaout"] << "Showing symmetric Hessian PDF uncertainty (--> ABM, (G)JR PDFs)." << endl;
      } else if ( chunc == "HA" ) {
         ePDFUnc = kHessianAsymmetric;
         shout["fnlo-tk-yodaout"] << "Showing asymmetric Hessian PDF uncertainty." << endl;
      } else if ( chunc == "HP" ) {
         ePDFUnc = kHessianAsymmetricMax;
         shout["fnlo-tk-yodaout"] << "Showing pairwise asymmetric Hessian PDF uncertainty (--> CTEQ|MSTW PDFs)." << endl;
      } else if ( chunc == "HC" ) {
         ePDFUnc = kHessianCTEQCL68;
         shout["fnlo-tk-yodaout"] << "Showing pairwise asymmetric Hessian PDF uncertainty rescaled to CL68 (--> CTEQ PDFs)." << endl;
      } else if ( chunc == "MC" ) {
         ePDFUnc = kMCSampling;
         shout["fnlo-tk-yodaout"] << "Showing MC sampling PDF uncertainty (--> NNPDF PDFs)." << endl;
      } else if ( chunc == "L6" ) {
         ePDFUnc = kLHAPDF6;
         shout["fnlo-tk-yodaout"] << "Showing LHAPDF6 PDF uncertainty (--> LHAPDF6 PDFs)." << endl;
      } else if ( chunc == "AS" ) {
         eAsUnc = kAsGRV;
         shout["fnlo-tk-yodaout"] << "Showing a_s(M_Z) uncertainty with GRV evolution." << endl;
      } else if ( chunc == "ST" ) {
         eStatUnc = kStatInt;
         shout["fnlo-tk-yodaout"] << "Showing statistical uncertainty of x section calculation." << endl;
      } else {
         error["fnlo-tk-yodaout"] << "Illegal choice of uncertainty, " << chunc << ", aborted!" << endl;
         exit(1);
      }
   }

   //! --- Fixed-order choice
   ESMOrder eOrder = kNextToLeading;
   string chord;
   bool bonly = false;
   if (argc > 4) {
      chord = (const char*) argv[4];
   }
   if (argc <= 4 || chord == "_") {
      chord = "NLO";
      shout["fnlo-tk-yodaout"] << "No request given for fixed-order precision, using NLO." << endl;
   } else {
      if ( chord == "LO" ) {
         eOrder = kLeading;
         shout["fnlo-tk-yodaout"] << "Deriving LO cross sections for comparison." << endl;
      } else if ( chord == "NLO" ) {
         eOrder = kNextToLeading;
         shout["fnlo-tk-yodaout"] << "Deriving NLO cross sections for comparison." << endl;
      } else if ( chord == "NLO_only" ) {
         eOrder = kNextToLeading;
         bonly = true;
         shout["fnlo-tk-yodaout"] << "Experimental: Deriving NLO contribution only for comparison." << endl;
      } else if ( chord == "NNLO" ) {
         eOrder = kNextToNextToLeading;
         shout["fnlo-tk-yodaout"] << "Deriving NNLO cross sections for comparison." << endl;
      } else if ( chord == "NNLO_only" ) {
         eOrder = kNextToNextToLeading;
         bonly = true;
         shout["fnlo-tk-yodaout"] << "Experimental: Deriving NNLO contribution only for comparison." << endl;
      } else {
         error["fnlo-tk-yodaout"] << "Illegal choice of fixed-order precision, " << chord << ", aborted!" << endl;
         exit(1);
      }
   }

   //! --- Normalization
   string chnorm;
   if (argc > 5) {
      chnorm = (const char*) argv[5];
   }
   if (argc <= 5 || chnorm == "_" || chnorm == "no") {
      chnorm = "no";
      shout["fnlo-tk-yodaout"] << "Preparing unnormalized cross sections." << endl;
   } else {
      shout["fnlo-tk-yodaout"] << "Normalizing cross sections. " << endl;
   }

   //--- Scale choice (flex-scale tables only; ignored for fix-scale tables)
   string chflex;
   if (argc > 6) {
      chflex = (const char*) argv[6];
   }
   if (argc <= 6 || chflex == "_") {
      chflex = "kScale1";
      shout["fnlo-tk-yodaout"] << "Using default mur=muf=scale 1." << endl;
   } else {
      shout["fnlo-tk-yodaout"] << "Using scale definition "+chflex+"." << endl;
   }

   //--- Central scale factor (e.g. to change from pT to 2*pT or m12 to m12/2)
   double sclfac;
   if (argc > 7) {
      chtmp = (const char*) argv[7];
   }
   if (argc <= 7 || chtmp == "_") {
      sclfac = 1.0;
      shout["fnlo-tk-yodaout"] << "No request given for central scale factor, using unmodified central scale." << endl;
   } else {
      sclfac = atof(argv[7]);
      if ( sclfac < 0.1 || sclfac > 10. ) {
         snprintf(buffer, sizeof(buffer), "Illegal value for sclfac, aborting! sclfac = %f\n",sclfac);
         error["fnlo-tk-yodaout"] << buffer << endl;
         exit(1);
      } else {
         shout["fnlo-tk-yodaout"] << "Using central scale factor of sclfac = " << sclfac << endl;
      }
   }

   //! --- Nonperturbative correction
   string chnp;
   if (argc > 8) {
      chnp = (const char*) argv[8];
   }
   if (argc <= 8 || chnp == "_" || chnp == "no") {
      chnp = "no";
      shout["fnlo-tk-yodaout"] << "Do not apply nonperturbative corrections." << endl;
   } else {
      shout["fnlo-tk-yodaout"] << "Apply nonperturbative corrections if available." << endl;
   }

   //--- Set x axis rescale velue
   double xrescale;
   if (argc > 9) {
      chtmp = (const char*) argv[9];
   }
   if (argc <= 9 || chtmp == "_") {
      xrescale = 1;
      shout["fnlo-tk-yodaout"] << "No request given for x-axis rescale factor, using default value of unity." << endl;
   } else {
      xrescale = atof(argv[9]);
      shout["fnlo-tk-yodaout"] << "Using x-axis rescale factor: " << xrescale << endl;
   }

   //--- Set uncertainty rescale velue
   double dxsrescale;
   if (argc > 10) {
      chtmp = (const char*) argv[10];
   }
   if (argc <= 10 || chtmp == "_") {
      dxsrescale = 1;
      shout["fnlo-tk-yodaout"] << "No request given for uncertainty rescale factor, using default value of unity." << endl;
   } else {
      dxsrescale = atof(argv[10]);
      shout["fnlo-tk-yodaout"] << "Using uncertainty rescale factor: " << dxsrescale << endl;
   }

   //--- Report about changed verbosity level at the beginning
   if (!VerbosityLevel.empty()) {
      shout["fnlo-tk-yodaout"] << "Using verbosity level: " << VerbosityLevel << endl;
   }

   //! ---  Too many arguments
   if (argc > 12) {
      error["fnlo-tk-yodaout"] << "Too many arguments, aborting!" << endl;
      exit(1);
   }
   yell << _CSEPSC << endl << endl;
   //---  End of parsing arguments

   //! --- fastNLO initialisation, attach table
   fastNLOTable table = fastNLOTable(tablename);

   //! Print essential table information
   if (fVerbosity < toVerbosity()["WARNING"]) {
      table.PrintContributionSummary(0);
      table.Print(0);
   }

   //! Initialise a fastNLO reader instance
   //! Note: This also initializes the cross section to the LO/NLO one!
   fastNLOLHAPDF* fnlo = NULL;
   if ( chunc != "AS" ) {
      fnlo = new fastNLOLHAPDF(table,PDFFile,0);
   } else {
      //! Need to reopen here with Alphas interface instead of LHAPDF
      fnlo = new fastNLOAlphas(tablename,PDFFile,0);
   }

   //! Check on existence of LO (Id = -1 if not existing)
   int ilo   = fnlo->ContrId(kFixedOrder, kLeading);
   if (ilo < 0) {
      if ( bonly ) {
         warn["fnlo-tk-yodaout"] << "LO not found, output for NLO_only or NNLO_only is experimental ...!" << endl;
      } else {
         error["fnlo-tk-yodaout"] << "LO not found; full cross section can not be derived, aborted!" << endl;
         exit(1);
      }
   } else {
      info["fnlo-tk-yodaout"] << "The LO contribution has Id: " << ilo << endl;
      if ( bonly ) {
         fnlo->SetContributionON(kFixedOrder, ilo, false);
      } else {
         fnlo->SetContributionON(kFixedOrder, ilo, true);
      }
   }
   unsigned int nOrder = 1;
   //! Check on existence of NLO (Id = -1 if not existing)
   int inlo  = fnlo->ContrId(kFixedOrder, kNextToLeading);
   if (inlo < 0) {
      info["fnlo-tk-yodaout"] << "No NLO contribution found!" << endl;
      if ( eOrder == kNextToLeading || ((eOrder > kNextToLeading) & !bonly) ) {
         error["fnlo-tk-yodaout"] << "Requested NLO not found, aborted!" << endl;
         exit(1);
      }
   } else {
      info["fnlo-tk-yodaout"] << "The NLO contribution has Id: " << inlo << endl;
      if ( eOrder == kNextToLeading ) {
         fnlo->SetContributionON(kFixedOrder, inlo, true);
      } else if ( (eOrder > kNextToLeading) && !bonly ) {
         fnlo->SetContributionON(kFixedOrder, inlo, true);
      } else {
         fnlo->SetContributionON(kFixedOrder, inlo, false);
      }
      nOrder = 2;
   }
   //! Check on existence of NNLO (Id = -1 if not existing)
   int innlo = fnlo->ContrId(kFixedOrder, kNextToNextToLeading);
   if (innlo < 0) {
      info["fnlo-tk-yodaout"] << "No NNLO contribution found!" << endl;
      if ( eOrder == kNextToNextToLeading || ((eOrder > kNextToNextToLeading) & !bonly) ) {
         error["fnlo-tk-yodaout"] << "Requested NNLO not found, aborted!" << endl;
         exit(1);
      }
   } else {
      info["fnlo-tk-yodaout"] << "The NNLO contribution has Id: " << innlo << endl;
      if ( eOrder == kNextToNextToLeading ) {
         fnlo->SetContributionON(kFixedOrder, innlo, true);
      } else if ( (eOrder > kNextToNextToLeading) && !bonly ) {
         fnlo->SetContributionON(kFixedOrder, innlo, true);
      } else {
         fnlo->SetContributionON(kFixedOrder, innlo, false);
      }
      nOrder = 3;
   }
   string sOrder = _OrdName[kFixedOrder][nOrder-1];

   //! Check on existence of non-perturbative corrections from LO MC
   int inpc1 = fnlo->ContrId(kNonPerturbativeCorrection, kLeading);
   if (inpc1 > -1 && (chnp == "yes" || chnp == "np" ) ) {
      info["fnlo-tk-yodaout"] << "Found non-perturbative correction factors. Switch on." << endl;
      bool SetOn = fnlo->SetContributionON(kNonPerturbativeCorrection, inpc1, true);
      if (!SetOn) {
         error["fnlo-tk-yodaout"] << "NPC1 not found, nothing to be done!" << endl;
         error["fnlo-tk-yodaout"] << "This should have been caught before!" << endl;
         exit(1);
      }
   }

   //! Normalize?
   bool lNorm = false;
   if ( chnorm == "yes" || chnorm == "norm" ) {
      if ( fnlo->IsNorm() ) {
         lNorm = true;
      } else {
         error["fnlo-read"] << "Normalization requested but not defined for this table, aborted!" << endl;
         exit(1);
      }
   }

   //! --- Determine dimensioning of observable bins in table
   const int NDim = fnlo->GetNumDiffBin();
   if (NDim < 1 || NDim > 3) {
      error["fnlo-tk-yodaout"] << "Found " << NDim << "-dimensional observable binning in table." << endl;
      error["fnlo-tk-yodaout"] << "Only up to three dimensions currently possible, aborted!" << endl;
      exit(1);
   }

   //! --- Get all required info from table
   //! Get binning
   vector < pair < double, double > > bins = fnlo->GetObsBinsBounds(NDim-1);

   //! For flex-scale tables:
   //! Possibility to redefine primary scale Q for mu_r and mu_f from the up to two stored scales
   //! Default choice is the first scale via enum 'kScale1'
   if (fnlo->GetIsFlexibleScaleTable()) {
      if ( chflex == "kScale1" ) {
         fnlo->SetMuFFunctionalForm(kScale1);
         fnlo->SetMuRFunctionalForm(kScale1);
      } else if ( chflex == "kScale2" ) {
         fnlo->SetMuFFunctionalForm(kScale2);
         fnlo->SetMuRFunctionalForm(kScale2);
      } else if ( chflex == "scale12" ) {
         fnlo->SetMuFFunctionalForm(kScale2);
         fnlo->SetMuRFunctionalForm(kScale1);
      } else if ( chflex == "scale21" ) {
         fnlo->SetMuFFunctionalForm(kScale1);
         fnlo->SetMuRFunctionalForm(kScale2);
      } else if ( chflex == "kProd" ) {
         fnlo->SetMuFFunctionalForm(kProd);
         fnlo->SetMuRFunctionalForm(kProd);
      } else if ( chflex == "kQuadraticSum" ) {
         fnlo->SetMuFFunctionalForm(kQuadraticSum);
         fnlo->SetMuRFunctionalForm(kQuadraticSum);
      } else {
         error["fnlo-tk-yodaout"] << "Unknown scale choice " << chflex << ", aborted!" << endl;
      }
   }

   //! Re-calculate cross sections to potentially include the above-selected non-perturbative factors
   fnlo->CalcCrossSection();
   //! If required get uncertainties (only for additive perturbative contributions)
   XsUncertainty XsUnc;
   string LineName;
   string UncName;
   if ( chunc == "2P" || chunc == "6P" ) {
      XsUnc = fnlo->GetXsUncertainty(eScaleUnc, lNorm, sclfac);
      UncName = " # Relative scale uncertainties";
      LineName += "_dxscl";
   } else if ( chunc == "AS" ) {
      XsUnc = fnlo->GetXsUncertainty(eAsUnc, lNorm);
      UncName = " # Relative a_s(M_Z) uncertainties";
      LineName += "_dxa_s";
   } else if ( chunc == "ST" ) {
      XsUnc = fnlo->GetXsUncertainty(eStatUnc, lNorm);
      UncName = " # Relative statistical uncertainties";
      LineName += "_dxst";
   } else if ( chunc != "none" ) {
      XsUnc = fnlo->GetXsUncertainty(ePDFUnc, lNorm);
      UncName = " # Relative PDF uncertainties";
      LineName += "_dxpdf";
   } else {
      XsUnc = fnlo->GetXsUncertainty(kScaleNone, lNorm);
      UncName = " # Without uncertainties";
      LineName += "_dxnone";
   }
   UncName = UncName + " (" + chunc + " " + PDFFile + " " + sOrder + ")";
   fastNLOTools::PrintXSUncertainty(XsUnc, UncName);

   //! --- Get RivetID
   //!     For 2+-dimensions determine running number in Rivet plot name by spotting the capital letter in "RIVET_ID=" in the fnlo table
   //!     For inverted order, the Id starts with "-" after "/".
   int    invert_order = 1;
   size_t capital_pos  = 0;
   string RivetId = fnlo->GetRivetId();
   if (RivetId.empty()) {
      warn["fnlo-tk-yodaout"] << "No Rivet ID found in fastNLO Table, stopping here without YODA output!" << endl;
      exit(0);
   }

   size_t i0 = RivetId.find("/");
   if (i0 == std::string::npos) {
      warn["fnlo-tk-yodaout"] << "Found malformed Rivet ID without \"/\": " << RivetId << " Stopping here without YODA output!" <<  endl;
      exit(0);
   }

   if ( NDim > 1 ) {
      /// Check for inversion of histogram order
      if ( RivetId.substr(i0,2) == "/-" ) {
         invert_order = -1;
         RivetId.erase(i0+1,1);
      }
      for (size_t i = i0; i < RivetId.size(); i++) {
         /// Identify capital letter in "d01-x01-y01" type strings, memorize position and lower it
         if (isupper(RivetId[i])) {
            RivetId[i] = tolower(RivetId[i]);
            capital_pos = i;
            break;
         }
      }
      if (capital_pos == 0) {
         error["fnlo-tk-yodaout"] << "Rivet ID found in fastNLO table does not indicate the n-dim. histogram counter, aborted." << endl;
         exit(1);
      }
   }

   //! --- Naming the file (and the legend line!) according to calculation order, PDF, and uncertainty choice
   string PDFName  = PDFFile.substr(0, min(11,(int)PDFFile.size()) );
   string FileName = PDFName;
   if ( chord  != "_" ) {FileName = chord + "_" + PDFName;}
   if ( chunc  != "none" ) {FileName += "_" + chunc;}
   if ( chnorm != "_" ) {FileName += "_" + chnorm;}
   if ( chflex != "_" ) {FileName += "_" + chflex;}
   if ( chnp   != "_" ) {FileName += "_" + chnp;}
   LineName = chord + "_" + PDFName + LineName;

   //! --- YODA analysis object creation and storage
#ifdef WITH_YODA
   YODA::Writer & writer = YODA::WriterYODA::create();                          //! Create the writer for the yoda file
   vector< YODA::AnalysisObject * > aos;                                   //! Vector of pointers to each of multiple analysis objects
#endif
   size_t counter = atoi(RivetId.substr(capital_pos +1, 2).c_str());

   //! --- Initialize dimension bin and continuous observable bin counter
   unsigned int NDimBins[NDim];
   NDimBins[0] = fnlo->GetNDim0Bins();
   unsigned int iobs = 0;
   //! Vector of 2D scatter plots
#ifdef WITH_YODA
   vector<YODA::Scatter2D> plots;
#endif

   //! --- 1D
   if (NDim == 1) {
      //! Vectors to fill 2D scatter plot
      vector < double > x;
      vector < double > y;
      vector < double > exminus;
      vector < double > explus;
      vector < double > eyminus;
      vector < double > eyplus;
      //! Loop over bins in outer (1st) dimension
      for (unsigned int k =0 ; k<NDimBins[0] ; k++) {
         x.push_back((bins[iobs].second + bins[iobs].first)/2.0*xrescale);
         explus.push_back((bins[iobs].second - bins[iobs].first)/2.0*xrescale);
         exminus.push_back((bins[iobs].second - bins[iobs].first)/2.0*xrescale);
         y.push_back(XsUnc.xs[iobs]);
         eyplus.push_back(XsUnc.xs[iobs]*XsUnc.dxsu[iobs]*dxsrescale);
         eyminus.push_back(std::abs(XsUnc.xs[iobs]*XsUnc.dxsl[iobs])*dxsrescale);
         iobs++;
      }
#ifdef WITH_YODA
      /// Pointer in order not to be deleted after we exit the loop, so we can then save them into the yoda file
      YODA::Scatter2D * plot = new YODA::Scatter2D(x,y,exminus,explus,eyminus,eyplus,"/" + RivetId,LineName);
      /// Insert the plot pointer into the vector of analysis object pointers
      aos.push_back(plot);
#endif
   }
   //! --- 2D
   else if (NDim == 2) {
      //! Loop over bins in outer (1st) dimension
      int nhist = 0;
      for (unsigned int j=0; j<NDimBins[0]; j++) {
         //! Vectors to fill 2D scatter plot
         nhist++;
         vector < double > x;
         vector < double > y;
         vector < double > exminus;
         vector < double > explus;
         vector < double > eyminus;
         vector < double > eyplus;
         //! Loop over bins in inner (2nd) dimension
         NDimBins[1] = fnlo->GetNDim1Bins(j);
         for (unsigned int k = 0; k<NDimBins[1]; k++) {
            x.push_back((bins[iobs].second + bins[iobs].first)/2.0*xrescale);
            explus.push_back((bins[iobs].second - bins[iobs].first)/2.0*xrescale);
            exminus.push_back((bins[iobs].second - bins[iobs].first)/2.0*xrescale);
            y.push_back(XsUnc.xs[iobs]);
            eyplus.push_back(XsUnc.xs[iobs]*XsUnc.dxsu[iobs]*dxsrescale);
            eyminus.push_back(std::abs(XsUnc.xs[iobs]*XsUnc.dxsl[iobs])*dxsrescale);
            iobs++;
         }
         /// Derive histogram counter
         size_t ihist = (invert_order > 0) ? (counter + nhist - 1) : (counter - nhist + 1);
         if ( ihist == 0 || ihist > 99 ) {
            error["fnlo-tk-yodaout"] << "Rivet histogram counter out of range, aborted!" << endl;
            error["fnlo-tk-yodaout"] << "ihist = " << ihist << endl;
            exit(1);
         }
         /// Convert size_t into string for naming
         stringstream histno;
         histno << ihist;
         /// Replace counter part in RivetId by histno
         RivetId.replace(capital_pos +3 - histno.str().size(), histno.str().size(), histno.str());
#ifdef WITH_YODA
         /// Pointer in order not to be deleted after we exit the loop, so we can then save the plots into the yoda file
         YODA::Scatter2D * plot = new YODA::Scatter2D(x,y,exminus,explus,eyminus,eyplus,"/" + RivetId,LineName);
         /// Insert the plot pointer into the vector of analysis object pointers
         aos.push_back(plot);
#endif
      }
   }
   //! --- 3D
   else if (NDim == 3) {
      //! Loop over bins in outer (1st) dimension
      int nhist = 0;
      for (unsigned int j=0; j<NDimBins[0]; j++) {
         //! Loop over bins in middle (2nd) dimension
         NDimBins[1] = fnlo->GetNDim1Bins(j);
         for (unsigned int k = 0; k<NDimBins[1]; k++) {
            //! Vectors to fill 2D scatter plot
            nhist++;
            vector < double > x;
            vector < double > y;
            vector < double > exminus;
            vector < double > explus;
            vector < double > eyminus;
            vector < double > eyplus;
            //! Loop over bins in inner (3rd) dimension
            NDimBins[2] = fnlo->GetNDim2Bins(j,k);
            for (unsigned int l = 0; l<NDimBins[2]; l++) {
               x.push_back((bins[iobs].second + bins[iobs].first)/2.0*xrescale);
               explus.push_back((bins[iobs].second - bins[iobs].first)/2.0*xrescale);
               exminus.push_back((bins[iobs].second - bins[iobs].first)/2.0*xrescale);
               y.push_back(XsUnc.xs[iobs]);
               eyplus.push_back(XsUnc.xs[iobs]*XsUnc.dxsu[iobs]*dxsrescale);
               eyminus.push_back(std::abs(XsUnc.xs[iobs]*XsUnc.dxsl[iobs])*dxsrescale);
               iobs++;
            }
            /// Derive histogram counter
            size_t ihist = (invert_order > 0) ? (counter + nhist - 1) : (counter - nhist + 1);
            if ( ihist == 0 || ihist > 99 ) {
               error["fnlo-tk-yodaout"] << "Rivet histogram counter out of range, aborted!" << endl;
               error["fnlo-tk-yodaout"] << "ihist = " << ihist << endl;
               exit(1);
            }
            /// Convert size_t into string for naming
            stringstream histno;
            histno << ihist;
            /// Replace counter part in RivetId by histno
            RivetId.replace(capital_pos +3 - histno.str().size(), histno.str().size(), histno.str());
#ifdef WITH_YODA
            /// Pointer in order not to be deleted after we exit the loop, so we can then save the plots into the yoda file
            YODA::Scatter2D * plot = new YODA::Scatter2D(x,y,exminus,explus,eyminus,eyplus,"/" + RivetId,LineName);
            /// Insert the plot pointer into the vector of analysis object pointers
            aos.push_back(plot);
#endif
         }
      }
   }

   //! --- Output
   //! Save histograms into the yoda file
#ifdef WITH_YODA
   writer.write( FileName + ".yoda", aos );
#endif
   yell << "" << endl;
   yell << _CSEPSC << endl;
   yell << " #" << endl;
#ifdef WITH_YODA
   shout << FileName + ".yoda was successfully produced" << endl;
#else
   shout << "Compiled without YODA support ==> No YODA file produced!" << endl;
#endif
   yell << " #" << endl;
   yell << _CSEPSC << endl << endl;
}
